"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>.
"""
from pyzora.enums import *
from pyzora.secret import *
from pyzora.game_secret import *
from pyzora.memory_secret import *
from pyzora.ring_secret import *
from pyzora.ring_types import *
from pyzora.child_behaviour_tools import *
__version__ = "1.0.0"
__author__ = "fortwoone"
